package net.sf.sido.spring.mvc;

import net.sf.sido.*;
import net.sf.sido.io.DataObjectReader;
import net.sf.sido.io.json.JSONDataObjectReader;
import net.sf.sido.io.xml.XMLDataObjectReader;
import org.springframework.core.MethodParameter;
import org.springframework.web.bind.support.WebArgumentResolver;
import org.springframework.web.context.request.NativeWebRequest;

import javax.servlet.ServletInputStream;
import javax.servlet.ServletRequest;
import java.util.Map;

public abstract class AbstractSiDOArgumentResolver implements WebArgumentResolver {

    @Override
    public Object resolveArgument(MethodParameter methodParameter, NativeWebRequest webRequest) throws Exception {
        if (supportsParameterType(methodParameter.getParameterType())) {
            return transform(resolveInternal(methodParameter, webRequest), methodParameter, webRequest);
        } else {
            return UNRESOLVED;
        }
    }

    protected abstract Object transform(DataObject dataObject, MethodParameter methodParameter, NativeWebRequest webRequest);

    protected DataObject resolveInternal(MethodParameter methodParameter, NativeWebRequest webRequest) throws Exception {
        ServletRequest req = webRequest.getNativeRequest(ServletRequest.class);
        // JSON content
        String contentType = req.getContentType();
        if (contentType != null && contentType.startsWith("sido/json")) {
            return resolveArgumentFromInputStream(methodParameter, webRequest, new JSONDataObjectReader());
        }
        // XML content
        else if (contentType != null && contentType.startsWith("sido/xml")) {
            return resolveArgumentFromInputStream(methodParameter, webRequest, new XMLDataObjectReader());
        }
        // Parameters
        else {
            return resolveArgumentFromParameters(methodParameter, webRequest);
        }
    }

    protected <T> DataObject resolveArgumentFromInputStream(MethodParameter methodParameter,
                                                            NativeWebRequest webRequest,
                                                            DataObjectReader<T> reader) throws Exception {
        ServletRequest req = webRequest.getNativeRequest(ServletRequest.class);
        ServletInputStream in = req.getInputStream();
        try {
            // Resolves the expected type
            DataType dataType = getType(methodParameter, webRequest, false);
            // Reads the object
            DataObject o = reader.read(in, getDataFactory(), dataType);
            // Checks the created object
            check(o, methodParameter);
            // OK
            return o;
        } finally {
            in.close();
        }
    }

    protected DataObject resolveArgumentFromParameters(MethodParameter methodParameter, NativeWebRequest webRequest) throws Exception {
        // Resolves the expected type
        DataType dataType = getType(methodParameter, webRequest, true);
        // Creates the instance
        DataObject o = dataType.newInstance();
        // Initialises the type?
        init(o, methodParameter);
        // For all parameters
        Map<String, String[]> parameters = webRequest.getParameterMap();
        for (Map.Entry<String, String[]> entry : parameters.entrySet()) {
            String name = entry.getKey();
            String[] values = entry.getValue();
            Property<Object> property = dataType.getProperty(name);
            if (property != null) {
                // Sets the value from a string
                o.setAsString(name, values);
            }
        }
        // Checks the created object
        check(o, methodParameter);
        // OK
        return o;
    }

    protected abstract void init(DataObject o, MethodParameter methodParameter);

    protected abstract void check(DataObject o, MethodParameter methodParameter);

    protected void validate(DataObject o, MethodParameter methodParameter) {
        ValidationResults results = o.validate();
        if (!results.isValid()) {
            throw new IllegalStateException(String.format("Cannot validate DataObject %s: %s", methodParameter.getParameterName(), results.getErrors()));
        }
    }

    protected DataType getType(MethodParameter methodParameter, NativeWebRequest webRequest, boolean typeRequired) throws Exception {
        String dataTypeName = getExpectedDataTypeName(methodParameter, webRequest, typeRequired);
        return dataTypeName != null ? getTypeFromName(dataTypeName) : null;
    }

    protected DataType getTypeFromName(String dataTypeName) {
        return getDataFactory().getMandatoryType(dataTypeName);
    }

    protected abstract String getExpectedDataTypeName(MethodParameter methodParameter, NativeWebRequest webRequest, boolean typeRequired) throws Exception;

    protected abstract DataFactory getDataFactory();

    protected abstract boolean supportsParameterType(Class<?> parameterType);
}
